@@ -48,6 +48,9 @@ class NBeats(BaseModel):
         the most
         important benchmark for univariate time series forecasting.
 
+        The :py:class:`~pytorch_forecasting.models.nhits.NHiTS` network has recently shown to consistently outperform
+        N-BEATS.
+
         Args:
             stack_types: One of the following values: “generic”, “seasonality" or “trend". A list of strings
                 of length 1 or ‘num_stacks’. Default and recommended value
@@ -224,7 +227,7 @@ class NBeats(BaseModel):
         """
         log, out = super().step(x, y, batch_idx=batch_idx)
 
-        if self.hparams.backcast_loss_ratio > 0:  # add loss from backcast
+        if self.hparams.backcast_loss_ratio > 0 and not self.predicting:  # add loss from backcast
             backcast = out["backcast"]
             backcast_weight = (
                 self.hparams.backcast_loss_ratio * self.hparams.prediction_length / self.hparams.context_length
@@ -236,8 +239,20 @@ class NBeats(BaseModel):
             else:
                 backcast_loss = self.loss(backcast, x["encoder_target"]) * backcast_weight
             label = ["val", "train"][self.training]
-            self.log(f"{label}_backcast_loss", backcast_loss, on_epoch=True, on_step=self.training)
-            self.log(f"{label}_forecast_loss", log["loss"], on_epoch=True, on_step=self.training)
+            self.log(
+                f"{label}_backcast_loss",
+                backcast_loss,
+                on_epoch=True,
+                on_step=self.training,
+                batch_size=len(x["decoder_target"]),
+            )
+            self.log(
+                f"{label}_forecast_loss",
+                log["loss"],
+                on_epoch=True,
+                on_step=self.training,
+                batch_size=len(x["decoder_target"]),
+            )
             log["loss"] = log["loss"] * forecast_weight + backcast_loss
 
         self.log_interpretation(x, out, batch_idx=batch_idx)
